/***************************************************************************/
/*   Este programa genera un flujo de k procesos, con k introducido por el */
/* usuario. Los procesos tienen una distrubucin de tiempos de llegada del */
/* tipo Poisson(1/3.46q).                                                  */
/*   Bsicamente el programa genera un nmero aleatorio entre 0 y 1 y a    */
/* partir de l se determina el tiempo transcurrido x entre la llegada del */
/* anterior proceso y este. Los instantes de tiempo en que llegan los      */
/* procesos sern t1 = 0 + x1, t2 = t1 + x2, t3 = t2 + x3,...              */
/*   Se desea que el nmero de procesos cortos sea del 95% y el de largos  */
/* del 5%. Un proceso largo tendr una duracin de rfaga de CPU = 4q+R*(96q-4q) */
/* y uno corto = 0.5q + R * (1.5q - 0.5q) con R  [0,1].                   */
/*   Los procesos sern almacenados en un fichero de texto donde el primer */
/* valor indica el tiempo de llegada y el segundo la duracin.             */
/***************************************************************************/
#include <stdio.h>   /* scanf, printf, fopen, fprintf, fclose */
#include <stdlib.h>  /* exit, random */
#include <math.h>    /* log (equivalente ln) */

#define l (1/3.46)   /* parmetro de la distribucin de Poisson */

/* Variables globales */
FILE *fichero;		/* fichero de procesos */

/* Prototipos */
void Abrir(char *s);
void Generar(void);

/***************************************************************************/
/* Funcin que crea un fichero de alto nivel para almacenar los procesos.  */
/***************************************************************************/
void Abrir(char *s) {

  fichero = fopen(s, "w+");
  if (fichero == NULL) {
    printf("\nERROR, el fichero %s no se ha podido crear.", s);
    exit(1);
  }
}

/***************************************************************************/
void Generar(void) {
unsigned long int k;	    /* n de procesos a generar */
register unsigned long int i;            /* variable para el bucle */
float t;		    /* tiempo de llegada del proceso actual */
float t_anterior;	    /* tiempo de llegada del proceso anterior */
float x;		    /* tiempo entre llegadas */
float aleatorio;	    /* n aleatorio */
float duracion;		    /* duracin del proceso */
unsigned int semilla;

  printf("\nIntroduce el nmero de procesos a generar: ");
  scanf("%ld", &k);
  printf("\nIntroduce la semilla: ");
  scanf("%d", &semilla);
  t_anterior = 0;
  srand(semilla);
  for (i = 1; i <= k; i++) {
    if ((i % 1000) == 0)
      printf("\nGenerando proceso %ld", i);
    /* tiempo de llegada del proceso */
    aleatorio = (float)(rand() % 100) / 100;  /* [0,1) */
    x = ((float)(-1/l))*log(1-aleatorio);
    t = t_anterior + x;

    /* duracin del proceso */
    aleatorio = (float)(rand() % 101) / 100;
    if (aleatorio < 0.95) {		/* procesos cortos */
      aleatorio = (float)(rand() % 101) / 100;
      duracion = 0.5 + aleatorio;
    }
    else {				/* procesos largos */
      aleatorio = (float)(rand() % 101) / 100;
      duracion = 4 + aleatorio * 92;
    }
    fprintf(fichero, "%f %f\n", t, duracion);
    t_anterior = t;
  }
  fclose(fichero);
}

/***************************************************************************/
int main(int argc, char *argv[]) {

  if (argc != 2){
    printf("\nERROR, forma de uso: genera archivo de procesos");
    exit(1);
  }
  Abrir(argv[1]);
  Generar();

  return 0;
}
