#include <stdio.h>
#include <stdlib.h>	/* malloc */
#include <math.h>       /* floor */

/* Tipos */
typedef enum {FALSE, TRUE} boolean;

typedef struct tipo_nodo {
  float llegada,		/* tiempo de llegada */
	duracion,		/* duracin que le queda */
	duracion_inicial;       /* duracion inicial */
  float quantum;	/* quantum de tiempo */
  struct tipo_nodo *sig;
};

typedef struct tipo_nodo nodo;

typedef struct tipo_cola {
  nodo *frente,		/* principio de la cola */
       *atras;          /* final de la cola */
};

typedef struct tipo_cola COLA;

typedef struct registro {
  float espera;			/* tiempo de espera */
  unsigned long long_cola,	/* longitud de la cola */
		contador,       /* n de elementos para tiempo de espera */
		contador2;	/* n de elementos para longitud de la cola */
};

typedef struct registro reg;

/* Variables */
unsigned long longitud;	  /* longitud de la cola */
COLA *c = NULL;
FILE *fichero;		  /* fichero de procesos */
FILE *salida;	          /* fichero de salida de long. de cola */
FILE *salida2;		  /* fichero de salida de tiempos de espera procesos cortos */
FILE *salida4;		  /* fichero de salida de tiempos espera procesos largos */
FILE *salida3;		  /* fichero de salida de longitud media de la cola */
float t_actual;           /* tiempo actual de planificacin */
float t_comienzo;         /* tiempo en que comenz a ejcutarse
			      el proceso actual */
unsigned short int cte;	  /* variable de compactacin */
nodo *n,           /* proceso que llega */
     *n_actual;    /* proceso en ejcucin */
reg vector[34];	   /* vector de datos para calcular la media de los tiempos
		      espera y la longitud de cola */
float cota_inferior,  /* cotas inferior y superior del intervalo */
      cota_superior;  /* estacionario */
long cambios;	      /* cambios de contexto */


/* Prototipos */
boolean Cola_vacia(COLA *cola);
void initCola(COLA *cola);
void push(COLA *cola, nodo *e);
void pop(COLA *cola, nodo *e);
void Libera_cola(COLA *cola);
void Abrir(char *s);
void Iniciar(void);
void Leer(nodo *e);
void Escribir(unsigned short *a, float *b, unsigned long *c);
void Tiempos_espera(void);
void Cola_media(void);
void Analizar(void);
void Analizar2(void);
void FCFS(void);
void RR(void);
void MRR(void);

/**************************************************************************/
boolean Cola_vacia(COLA *cola) {
  if (cola -> frente == NULL)
    return TRUE;
  else
    return FALSE;
}

/**************************************************************************/
void initCola(COLA *cola) {
register unsigned i;

  cola -> frente = NULL;
  cola -> atras = NULL;
  longitud = 0;
  t_actual = 0;

  for (i = 0; i < 34; i++) {
    vector[i].espera = 0;
    vector[i].contador = 0;
    vector[i].long_cola = 0;
    vector[i].contador2 = 0;
  }
}

/**************************************************************************/
void push(COLA *cola, nodo *e) {
nodo *q;

  q = (nodo *) malloc(sizeof(nodo));
  if (q == NULL) {
    printf("\nERROR, no se ha podido dar memoria al nodo.");
    exit(1);
  }
  *q = *e;
  q -> sig = NULL;
  if (cola -> frente ==NULL)
    cola -> frente = q;
  else
    cola -> atras -> sig = q;
  cola -> atras = q;
}

/**************************************************************************/
void pop(COLA *cola, nodo *e) {
nodo *q;

  if (Cola_vacia(cola)) {
    printf("\nERROR, cola vaca.");
    exit(1);
  }
  *e = *(cola -> frente);
  q = cola -> frente;
  cola -> frente = cola -> frente -> sig;
  free(q);
}

/**************************************************************************/
void Libera_cola(COLA *cola) {
nodo *q;

  while (!Cola_vacia(cola)) {
    q = cola -> frente;
    cola -> frente = cola -> frente -> sig;
    free(q);
  }
}

/***************************************************************************/
/* Funcin que abre un fichero de alto nivel para leer los procesos.       */
/***************************************************************************/
void Abrir(char *s) {

  fichero = fopen(s, "r");
  if (fichero == NULL) {
    printf("\nERROR, el fichero %s no se ha podido abrir.", s);
    exit(1);
  }
}

/***************************************************************************
  Funcin que da memoria a la cola, a los nodos auxiliares y pregunta por
 el nivel de compactacin de los datos.
***************************************************************************/
void Iniciar(void) {

  /* memoria para la cola */
  c = (COLA *) malloc(sizeof(COLA));
  if (c == NULL) {
    printf("\nERROR, no se ha podido dar memoria a la cola.");
    exit(1);
  }

  /* memoria para los nodos auxiliares */
  n_actual = (nodo *) malloc(sizeof(nodo));
  if (n_actual == NULL) {
    printf("\nERROR, no se ha podido dar memoria al nodo.");
    exit(1);
  }
  n = (nodo *) malloc(sizeof(nodo));
  if (n == NULL) {
    printf("\nERROR, no se ha podido dar memoria al nodo.");
    exit(1);
  }

  printf("\nNivel de compactacin: ");
  scanf("%d", &cte);

  printf("Cota inferior de intervalo estacionario: ");
  scanf("%f", &cota_inferior);
  printf("Cota superior de intervalo estacionario: ");
  scanf("%f", &cota_superior);
}

/***************************************************************************/
void Leer(nodo *e) {
float l, d;		/* variables auxiliares */

  fscanf(fichero, "%f %f\n", &l, &d);
  e->llegada = l;
  e->duracion = d;
  e->duracion_inicial = d;
  e->quantum = 1;
}

/***************************************************************************
 Parmetros:
 a - contador para la compactacin
 b - suma de los tiempos de llegada
 c - suma de las longitudes de cola
***************************************************************************/
void Escribir(unsigned short *a, float *b, unsigned long *c) {

  (*a)++;
  *b += t_actual;
  *c += longitud;
  if (*a == cte) {
    fprintf(salida,"%f %ld\n", (*b / cte), (long)(*c / cte));
    *a = 0;
    *b = 0;
    *c = 0;
  }
}

/***************************************************************************
  Funcin que almacena los tiempos de espera en cada intervalo de duracin
 de proceso. Estos tiempos se guardan si el proceso actual est en el
 intervalo estacionario.
***************************************************************************/
void Tiempos_espera(void) {
float t_espera;
int intervalo;

  if ((t_actual >= cota_inferior) && (t_actual < cota_superior)) {
    intervalo = ceil(n_actual->duracion_inicial / 4.0) - 1;
    t_espera = t_actual - n_actual->llegada;
    if ((intervalo == 0) && (n_actual->duracion_inicial != 4.0)) {
      intervalo = floor(n_actual->duracion_inicial * 10) - 6;
      if (intervalo == -1)
	intervalo = 0;
      vector[intervalo].espera += t_espera;
      if (n_actual->duracion_inicial == n_actual->duracion)
	vector[intervalo].contador++;
    }
    else {
      vector[intervalo+10].espera += t_espera;
      if (n_actual->duracion_inicial == n_actual->duracion)
	 vector[intervalo+10].contador++;
    }
  }
}

/***************************************************************************/
void Cola_media(void) {
short int intervalo;

  if ((t_actual >= cota_inferior) && (t_actual < cota_superior)) {
    intervalo = ceil(n_actual->duracion_inicial / 4.0) - 1;
    vector[intervalo].long_cola += longitud;
    vector[intervalo].contador2++;
  }
}


/***************************************************************************/
void Analizar(void) {
register unsigned i;
float media;			/* media tiempos de espera */
unsigned long media2;		/* media de la cola */
unsigned long media_t=0;	/* media de la cola total */
unsigned long cont=0;

  /* Tiempos de espera para procesos cortos */
  for (i = 0; i < 10; i++) {
    if (vector[i].contador == 0)
      media = 0;
    else
      media = vector[i].espera / vector[i].contador;
    fprintf(salida2,"%d %f\n", i+1, media);
  }

  /* Tiempos de espera para procesos largos */
  for (i = 10; i < 34; i++) {
    if (vector[i].contador == 0)
      media = 0;
    else
      media = vector[i].espera / vector[i].contador;
    fprintf(salida4,"%d %f\n", i+1, media);
  }

  /* Longitud media de cola */
  for (i = 0; i < 24; i++) {
    if (vector[i].contador2 == 0)
      media2 = 0;
    else
      media2 = vector[i].long_cola / vector[i].contador2;
    fprintf(salida3,"%d %dl\n", i+1, media2);
    media_t += vector[i].long_cola;
    cont += vector[i].contador2;
  }
  printf("\nCambios de contexto: %ld", cambios);
  printf("\nLongitud media de cola: %ld", (media_t / cont));
}

/***************************************************************************/
void Tiempos_espera2(void) {
float t_espera;

  if ((t_actual >= cota_inferior) && (t_actual < cota_superior)) {
    t_espera = t_actual - n_actual->llegada;
    if (n_actual->duracion_inicial <= 1) {
      vector[0].espera += t_espera;
      if (n_actual->duracion_inicial == n_actual->duracion)
	vector[0].contador++;
    }
    else {
      if ((n_actual->duracion_inicial > 1) && (n_actual->duracion_inicial <= 3)) {
	vector[1].espera += t_espera;
	if (n_actual->duracion_inicial == n_actual->duracion)
	  vector[1].contador++;
      }
      else
	if ((n_actual->duracion_inicial > 3) && (n_actual->duracion_inicial <= 7)) {
	  vector[2].espera += t_espera;
	  if (n_actual->duracion_inicial == n_actual->duracion)
	    vector[2].contador++;
	}
	if ((n_actual->duracion_inicial > 7) && (n_actual->duracion_inicial <= 15)) {
	  vector[3].espera += t_espera;
	  if (n_actual->duracion_inicial == n_actual->duracion)
	    vector[3].contador++;
	}
	if ((n_actual->duracion_inicial > 15) && (n_actual->duracion_inicial <= 31)) {
	  vector[4].espera += t_espera;
	  if (n_actual->duracion_inicial == n_actual->duracion)
	    vector[4].contador++;
	}
	if ((n_actual->duracion_inicial > 31) && (n_actual->duracion_inicial <= 63)) {
	  vector[5].espera += t_espera;
	  if (n_actual->duracion_inicial == n_actual->duracion)
	    vector[5].contador++;
	}
	/* if ((n_actual->duracion_inicial > 44) && (n_actual->duracion_inicial <= 64)) {
	  vector[6].espera += t_espera;
	  if (n_actual->duracion_inicial == n_actual->duracion)
	    vector[6].contador++;
	} */
	if (n_actual->duracion_inicial > 63) {
	  vector[6].espera += t_espera;
	  if (n_actual->duracion_inicial == n_actual->duracion)
	    vector[6].contador++;
	}
    }
  }
}

/***************************************************************************
 Los intervalos son <=1, 1-3, 3-7, 7-15, 15-31, 31-63, >63
***************************************************************************/
void Analizar2(void) {
register unsigned i;
float media;			/* media tiempos de espera */
unsigned long media2;		/* media de la cola */
short int aux[7]={1,2,5,11,23,47,95};
unsigned long media_t=0;	/* media de la cola total */
unsigned long cont=0;


  /* Tiempos de espera */
  for (i = 0; i < 7; i++) {
    if (vector[i].contador == 0)
      media = 0;
    else
      media = vector[i].espera / vector[i].contador;
    fprintf(salida2,"%d %f\n", aux[i], media);
  }

  /* Longitud media de cola */
  for (i = 0; i < 24; i++) {
    if (vector[i].contador2 == 0)
      media2 = 0;
    else
      media2 = vector[i].long_cola / vector[i].contador2;
    fprintf(salida3,"%d %dl\n", i+1, media2);
    media_t += vector[i].long_cola;
    cont += vector[i].contador2;
  }
  printf("\nCambios de contexto: %ld", cambios);
  printf("\nLongitud media de cola: %ld", (media_t / cont));
}

/***************************************************************************
  Procedimiento que simula el planificador FCFS
***************************************************************************/
void FCFS(void) {
unsigned short int count = 0;	   /* contador para la compactacin */
float suma_lleg=0;
unsigned long suma_long=0;

  printf("\nPlanificando FCFS...");
  cambios = 0;
  /* posicionarse al principio de los procesos */
  fseek(fichero, 0L, SEEK_SET);

  /* apertura fichero salida para la cola */
  salida = fopen("colafcfs.dat", "w+");
  if (salida == NULL) {
    printf("\nERROR, el fichero de salida de la longitud de cola para FCFS no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los tiempos de espera procesos largos */
  salida4 = fopen("waitfcfs.lar", "w+");
  if (salida4 == NULL) {
    printf("\nERROR, el fichero de salida de los tiempos de espera (procesos largos) para FCFS no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los tiempos de espera procesos cortos */
  salida2 = fopen("waitfcfs.cor", "w+");
  if (salida2 == NULL) {
    printf("\nERROR, el fichero de salida de los tiempos de espera (procesos cortos) para FCFS no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los longitud media de cola */
  salida3 = fopen("lmcfcfs.dat", "w+");
  if (salida3 == NULL) {
    printf("\nERROR, el fichero de salida de la longitud media de cola para FCFS no se ha podido crear.");
    exit(1);
  }

  initCola(c);
  /* primer proceso */
  Leer(n_actual);
  cambios++;
  t_comienzo = t_actual = n_actual->llegada;
  Escribir(&count, &suma_lleg, &suma_long);

  /* resto de los procesos */
  while (!feof(fichero)) {
    Leer(n);
    /* el proceso en ejecucin acaba despus del que llega */
    if ((t_comienzo + n_actual->duracion) > (n->llegada))  {
      t_actual = n->llegada;
      push(c,n);
      Cola_media();
      longitud++;
      Escribir(&count, &suma_lleg, &suma_long);
    }
/* el proceso que llega lo hace despues de que acabe el que est en ejecucin */
    else {
      t_actual = t_comienzo + n_actual->duracion;  /* instante en que acaba */
      Escribir(&count, &suma_lleg, &suma_long);
      if (Cola_vacia(c)) {
	*n_actual = *n;
	cambios++;
	t_actual = t_comienzo= n->llegada;
	Escribir(&count, &suma_lleg, &suma_long);
      }
      else {
	pop(c,n_actual);
	cambios++;
	longitud--;
	t_comienzo = t_actual;
	Tiempos_espera();
/* el proceso que sale de la cola puede acabar antes del que llega */
	while (((t_actual + n_actual->duracion) <= (n->llegada)) &&
	       (!feof(fichero))) {
	  t_actual += n_actual->duracion;
	  Escribir(&count, &suma_lleg, &suma_long);
	  if (Cola_vacia(c)) {
	    *n_actual = *n;
	    cambios++;
	    t_actual = t_comienzo= n->llegada;
	    Escribir(&count, &suma_lleg, &suma_long);
	    Leer(n);
	  }
	  else {
	    pop(c,n_actual);
	    cambios++;
	    longitud--;
	    t_comienzo = t_actual;
	    Tiempos_espera();
	    Escribir(&count, &suma_lleg, &suma_long);
	  }
	} /* end while */
	/* comparacin del ltimo proceso con el actual */
	if (feof(fichero))
	  if ((t_actual + n_actual->duracion) > (n->llegada)) {
	    /* el ltimo llega antes de que acabe */
	    t_actual = n->llegada;
	    push(c,n);
	    Cola_media();
	    longitud++;
	    Escribir(&count, &suma_lleg, &suma_long);
	  }
	  else {
	    /* el ltimo llega despus de que acabe */
	    t_actual += n_actual->duracion;
	    Escribir(&count, &suma_lleg, &suma_long);
	    t_actual = n->llegada;
	    Escribir(&count, &suma_lleg, &suma_long);
	    t_actual += n->duracion;
	    Escribir(&count, &suma_lleg, &suma_long);
	  }
	else {
	  t_actual = n->llegada;
	  push(c,n);
	  Cola_media();
	  longitud++;
	  Escribir(&count, &suma_lleg, &suma_long);
	}
      } /* end else */
    } /* end else */
  } /* end while (!feof(fichero)) */
  /* vaciamos la cola de preparados */
  while (!Cola_vacia(c)) {
    t_actual = t_comienzo+ n_actual->duracion;
    Escribir(&count, &suma_lleg, &suma_long);
    pop(c,n_actual);
    cambios++;
    t_comienzo = t_actual;
    longitud--;
    Tiempos_espera();
  }

  Libera_cola(c);
  Analizar();
  fclose(salida);
  fclose(salida2);
  fclose(salida3);
  fclose(salida4);
}

/***************************************************************************
 Planificador Round Robin.
***************************************************************************/
void RR(void) {
unsigned short int count = 0;	   /* contador para la compactacin */
float suma_lleg=0;
unsigned long suma_long=0;

  printf("\n\nPlanificando RR...");
  cte = 2 * cte;
  cambios = 0;
  /* posicionarse al principio de los procesos */
  fseek(fichero, 0L, SEEK_SET);

  /* apertura fichero salida */
  salida = fopen("colarr.dat", "w+");
  if (salida == NULL) {
    printf("\nERROR, el fichero de salida de la longitud de cola para RR no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los tiempos de espera procesos largos */
  salida4 = fopen("waitrr.lar", "w+");
  if (salida4 == NULL) {
    printf("\nERROR, el fichero de salida de los tiempos de espera (procesos largos) para RR no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los tiempos de espera procesos cortos */
  salida2 = fopen("waitrr.cor", "w+");
  if (salida2 == NULL) {
    printf("\nERROR, el fichero de salida de los tiempos de espera (procesos cortos) para RR no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los longitud media de cola */
  salida3 = fopen("lmcrr.dat", "w+");
  if (salida3 == NULL) {
    printf("\nERROR, el fichero de salida de la longitud media de cola para RR no se ha podido crear.");
    exit(1);
  }

  initCola(c);
  /* primer proceso */
  Leer(n_actual);
  cambios++;
  t_comienzo = t_actual = n_actual->llegada;
  Escribir(&count, &suma_lleg, &suma_long);
  Leer(n);
  /* resto de los procesos */
  while (!feof(fichero)) {
    /* el proceso en ejecucin acaba despus del que llega */
    if ((t_comienzo + n_actual->duracion) > (n->llegada))
      if (n_actual->duracion > n_actual->quantum) {
	if (n->llegada < (t_comienzo+n_actual->quantum)) {
	  while (((t_comienzo + n_actual->quantum) > (n->llegada)) &&
	       (!feof(fichero))) {
	    t_actual = n->llegada;
	    push(c,n);
	    Cola_media();
	    longitud++;
	    Leer(n);
	    Escribir(&count, &suma_lleg, &suma_long);
	  }
	  if (feof(fichero))
	    if ((t_comienzo + n_actual->quantum) > n->llegada) {
	      t_actual = n->llegada;
	      push(c,n);
	      Cola_media();
	      longitud++;
	      Escribir(&count, &suma_lleg, &suma_long);
	    }
	    else {
	      t_actual = t_comienzo + n_actual->quantum;
	      n_actual->llegada = t_actual;
	      n_actual->duracion -= n_actual->quantum;
	      push(c,n_actual);
	      Cola_media();
	      pop(c,n_actual);
	      cambios++;
	      Tiempos_espera();
	      Escribir(&count, &suma_lleg, &suma_long);
	      t_comienzo= t_actual;
	    }
	  else {
	    t_actual = t_comienzo + n_actual->quantum;
	    n_actual->llegada = t_actual;
	    n_actual->duracion -= n_actual->quantum;
	    push(c,n_actual);
	    Cola_media();
	    pop(c,n_actual);
	    cambios++;
	    Tiempos_espera();
	    Escribir(&count, &suma_lleg, &suma_long);
	    t_comienzo= t_actual;
	  }
	}
	else {
	  t_comienzo = t_actual += n_actual -> quantum;
	  n_actual->duracion -= n_actual->quantum;
	  n_actual->llegada = t_actual;
	  push(c,n_actual);
	  Cola_media();
	  Escribir(&count, &suma_lleg, &suma_long);
	  pop(c,n_actual);
	  cambios++;
	  Tiempos_espera();
	}
      }
      else { /* termina antes de que se le acabe el quantum y el nuevo llega antes */
	t_actual = n->llegada;
	push(c,n);
	Cola_media();
	longitud++;
	Escribir(&count, &suma_lleg, &suma_long);
	Leer(n);
      }
/* el proceso que llega lo hace depues de que acabe el que est en ejecucin */
    else {
      if (n_actual->duracion > n_actual->quantum) {
	t_comienzo = t_actual += n_actual->quantum;
	n_actual->llegada = t_actual;
	n_actual->duracion -= n_actual->quantum;
	push(c,n_actual);
	Cola_media();
	Escribir(&count, &suma_lleg, &suma_long);
	pop(c,n_actual);
	cambios++;
	Tiempos_espera();
      }
      else { /* acaba antes de que se le termine el quantum */
	t_actual = t_comienzo + n_actual->duracion;  /* instante en que acaba */
	Escribir(&count, &suma_lleg, &suma_long);
	if (Cola_vacia(c)) {
	  *n_actual = *n;
	  cambios++;
	  t_actual = t_comienzo= n->llegada;
	  Escribir(&count, &suma_lleg, &suma_long);
	  Leer(n);
	}
	else {
	  pop(c,n_actual);
	  cambios++;
	  longitud--;
	  t_comienzo = t_actual;
	  Tiempos_espera();
	  /* el proceso que sale de la cola puede acabar antes del que llega */
	  while (((t_actual + n_actual->duracion) <= (n->llegada)) &&
	       (!feof(fichero))) {
	    if (n_actual->duracion > n_actual->quantum) {
	      t_comienzo = t_actual += n_actual->quantum;
	      n_actual->duracion -= n_actual->quantum;
	      n_actual->llegada = t_actual;
	      push(c,n_actual);
	      Cola_media();
	      Escribir(&count, &suma_lleg, &suma_long);
	      pop(c,n_actual);
	      cambios++;
	      Tiempos_espera();
	    }
	    else {
	      t_actual += n_actual->duracion;
	      Escribir(&count, &suma_lleg, &suma_long);
	      if (Cola_vacia(c)) {
		*n_actual = *n;
		cambios++;
		t_actual = t_comienzo= n->llegada;
		Escribir(&count, &suma_lleg, &suma_long);
		Leer(n);
	      }
	      else {
		pop(c,n_actual);
		cambios++;
		longitud--;
		t_comienzo = t_actual;
		Escribir(&count, &suma_lleg, &suma_long);
		Tiempos_espera();
	      }
	    }
	  } /* end while */
	  /* comparacin del ltimo proceso con el actual */
	  if (feof(fichero)) {
	    /* el ltimo llega antes de que acabe el actual */
	    if ((t_actual + n_actual->duracion) > (n->llegada)) {
	    /* le queda ms del quantum por terminar */
	      if (n_actual->duracion > n_actual->quantum) {
		if (n->llegada < (t_comienzo + n_actual->quantum)) {
		  t_actual = n->llegada;
		  push(c,n);
		  Cola_media();
		  longitud++;
		  Escribir(&count, &suma_lleg, &suma_long);
		}
		else {
		  t_comienzo = t_actual += n_actual -> quantum;
		  n_actual->duracion -= n_actual->quantum;
		  n_actual->llegada = t_actual;
		  push(c,n_actual);
		  Cola_media();
		  Escribir(&count, &suma_lleg, &suma_long);
		  pop(c,n_actual);
		  cambios++;
		  Tiempos_espera();
		}
	      }
	      else { /* le queda menos del quantum por terminar */
		t_actual = n->llegada;
		push(c,n);
		Cola_media();
		longitud++;
		Escribir(&count, &suma_lleg, &suma_long);
	      }
	    }
	    else { /* el ltimo llega despus de que acabe el actual */
	      if (n_actual->duracion > n_actual->quantum) {
		t_comienzo = t_actual += n_actual -> quantum;
		n_actual->duracion -= n_actual->quantum;
		n_actual->llegada = t_actual;
		push(c,n_actual);
		Cola_media();
		Escribir(&count, &suma_lleg, &suma_long);
		pop(c,n_actual);
		cambios++;
		Tiempos_espera();
	      }
	      else {
		t_actual += n_actual->duracion;
		Escribir(&count, &suma_lleg, &suma_long);
		if (Cola_vacia(c)) {
		  *n_actual = *n;
		  cambios++;
		  t_actual = t_comienzo= n->llegada;
		  Escribir(&count, &suma_lleg, &suma_long);
		}
		else {
		  pop(c,n_actual);
		  cambios++;
		  Tiempos_espera();
		  longitud--;
		  t_comienzo = t_actual;
		}
	      }
	    }
	  }
	} /* end else */
      } /* end else */
    } /* end else (proceso no dura ms de 1) */
  } /* end while (!feof(fichero)) */

  /* vaciamos la cola de preparados */
  while (!Cola_vacia(c)) {
    if (n_actual->duracion > n_actual->quantum) {
      t_actual = t_comienzo += n_actual->quantum;
      n_actual->duracion -= n_actual->quantum;
      n_actual->llegada = t_actual;
      push(c,n_actual);
      Cola_media();
      Escribir(&count, &suma_lleg, &suma_long);
      pop(c,n_actual);
      cambios++;
      Tiempos_espera();
    }
    else {
      t_actual = t_comienzo+ n_actual->duracion;
      Escribir(&count, &suma_lleg, &suma_long);
      pop(c,n_actual);
      cambios++;
      Tiempos_espera();
      t_comienzo = t_actual;
      longitud--;
    }
  }

   /* terminamos el ltimo */
  t_actual = t_comienzo + n_actual->duracion;
  Escribir(&count, &suma_lleg, &suma_long);


  Libera_cola(c);
  Analizar();
  fclose(salida);
  fclose(salida2);
  fclose(salida3);
  fclose(salida4);
}

/***************************************************************************
 Planificador Round Robin Modificado.
***************************************************************************/
void MRR(void) {
unsigned short int count = 0;	   /* contador para la compactacin */
float suma_lleg=0;
unsigned long suma_long=0;

  printf("\n\nPlanificandoo MRR...");
  cte = cte / 2;
  cte = cte + floor((cte / 2));
  cambios = 0;
  /* posicionarse al principio de los procesos */
  fseek(fichero, 0L, SEEK_SET);

  /* apertura fichero salida */
  salida = fopen("colamrr.dat", "w+");
  if (salida == NULL) {
    printf("\nERROR, el fichero de salida de la longitud de cola para MRR no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los tiempos de espera */
  salida2 = fopen("waitmrr.dat", "w+");
  if (salida2 == NULL) {
    printf("\nERROR, el fichero de salida de los tiempos de espera para MRR no se ha podido crear.");
    exit(1);
  }
  /* apertura fichero salida para los tiempos de espera procesos largos */
/*  salida4 = fopen("waitmrr.lar", "w+");
  if (salida4 == NULL) {
    printf("\nERROR, el fichero de salida de los tiempos de espera (procesos largos) para MRR no se ha podido crear.");
    exit(1);
  }*/
  /* apertura fichero salida para los longitud media de cola */
  salida3 = fopen("lmcmrr.dat", "w+");
  if (salida3 == NULL) {
    printf("\nERROR, el fichero de salida de la longitud media de cola para MRR no se ha podido crear.");
    exit(1);
  }

  initCola(c);
  /* primer proceso */
  Leer(n_actual);
  cambios++;
  t_comienzo = t_actual = n_actual->llegada;
  Escribir(&count, &suma_lleg, &suma_long);
  Leer(n);
  /* resto de los procesos */
  while (!feof(fichero)) {
    /* el proceso en ejecucin acaba despus del que llega */
    if ((t_comienzo + n_actual->duracion) > (n->llegada))
      if (n_actual->duracion > n_actual->quantum) {
	if (n->llegada < (t_comienzo+n_actual->quantum)) {
	  while (((t_comienzo + n_actual->quantum) > (n->llegada)) &&
	       (!feof(fichero))) {
	    t_actual = n->llegada;
	    push(c,n);
	    Cola_media();
	    longitud++;
	    Leer(n);
	    Escribir(&count, &suma_lleg, &suma_long);
	  }
	  if (feof(fichero))
	    if ((t_comienzo + n_actual->quantum) > n->llegada) {
	      t_actual = n->llegada;
	      push(c,n);
	      Cola_media();
	      longitud++;
	      Escribir(&count, &suma_lleg, &suma_long);
	    }
	    else {
	      t_actual = t_comienzo + n_actual->quantum;
	      n_actual->llegada = t_actual;
	      n_actual->duracion -= n_actual->quantum;
	      n_actual->quantum = 2 * n_actual->quantum;
	      push(c,n_actual);
	      Cola_media();
	      pop(c,n_actual);
	      cambios++;
	      Tiempos_espera2();
	      Escribir(&count, &suma_lleg, &suma_long);
	      t_comienzo= t_actual;
	    }
	  else {
	    t_actual = t_comienzo + n_actual->quantum;
	    n_actual->llegada = t_actual;
	    n_actual->duracion -= n_actual->quantum;
	    n_actual->quantum = 2 * n_actual->quantum;
	    push(c,n_actual);
	    Cola_media();
	    pop(c,n_actual);
	    cambios++;
	    Tiempos_espera2();
	    Escribir(&count, &suma_lleg, &suma_long);
	    t_comienzo= t_actual;
	  }
	}
	else {
	  t_comienzo = t_actual += n_actual -> quantum;
	  n_actual->duracion -= n_actual->quantum;
	  n_actual->llegada = t_actual;
	  n_actual->quantum = 2 * n_actual->quantum;
	  push(c,n_actual);
	  Cola_media();
	  Escribir(&count, &suma_lleg, &suma_long);
	  pop(c,n_actual);
	  cambios++;
	  Tiempos_espera2();
	}
      }
      else { /* termina antes de que se le acabe el quantum y el nuevo llega antes */
	t_actual = n->llegada;
	push(c,n);
	Cola_media();
	longitud++;
	Escribir(&count, &suma_lleg, &suma_long);
	Leer(n);
      }
/* el proceso que llega lo hace depues de que acabe el que est en ejecucin */
    else {
      if (n_actual->duracion > n_actual->quantum) {
	t_comienzo = t_actual += n_actual->quantum;
	n_actual->llegada = t_actual;
	n_actual->duracion -= n_actual->quantum;
	n_actual->quantum = 2 * n_actual->quantum;
	push(c,n_actual);
	Cola_media();
	Escribir(&count, &suma_lleg, &suma_long);
	pop(c,n_actual);
	cambios++;
	Tiempos_espera2();
      }
      else { /* acaba antes de que se le termine el quantum */
	t_actual = t_comienzo + n_actual->duracion;  /* instante en que acaba */
	Escribir(&count, &suma_lleg, &suma_long);
	if (Cola_vacia(c)) {
	  *n_actual = *n;
	  cambios++;
	  t_actual = t_comienzo= n->llegada;
	  Escribir(&count, &suma_lleg, &suma_long);
	  Leer(n);
	}
	else {
	  pop(c,n_actual);
	  cambios++;
	  longitud--;
	  t_comienzo = t_actual;
	  Tiempos_espera2();
	  /* el proceso que sale de la cola puede acabar antes del que llega */
	  while (((t_actual + n_actual->duracion) <= (n->llegada)) &&
	       (!feof(fichero))) {
	    if (n_actual->duracion > n_actual->quantum) {
	      t_comienzo = t_actual += n_actual->quantum;
	      n_actual->duracion -= n_actual->quantum;
	      n_actual->llegada = t_actual;
	      n_actual->quantum = 2 * n_actual->quantum;
	      push(c,n_actual);
	      Cola_media();
	      Escribir(&count, &suma_lleg, &suma_long);
	      pop(c,n_actual);
	      cambios++;
	      Tiempos_espera2();
	    }
	    else {
	      t_actual += n_actual->duracion;
	      Escribir(&count, &suma_lleg, &suma_long);
	      if (Cola_vacia(c)) {
		*n_actual = *n;
		cambios++;
		t_actual = t_comienzo= n->llegada;
		Escribir(&count, &suma_lleg, &suma_long);
		Leer(n);
	      }
	      else {
		pop(c,n_actual);
		cambios++;
		Tiempos_espera2();
		longitud--;
		t_comienzo = t_actual;
		Escribir(&count, &suma_lleg, &suma_long);
	      }
	    }
	  } /* end while */
	  /* comparacin del ltimo proceso con el actual */
	  if (feof(fichero)) {
	    /* el ltimo llega antes de que acabe el actual */
	    if ((t_actual + n_actual->duracion) > (n->llegada)) {
	    /* le queda ms del quantum por terminar */
	      if (n_actual->duracion > n_actual->quantum) {
		if (n->llegada < (t_comienzo + n_actual->quantum)) {
		  t_actual = n->llegada;
		  push(c,n);
		  Cola_media();
		  longitud++;
		  Escribir(&count, &suma_lleg, &suma_long);
		}
		else {
		  t_comienzo = t_actual += n_actual -> quantum;
		  n_actual->duracion -= n_actual->quantum;
		  n_actual->llegada = t_actual;
		  n_actual->quantum = 2 * n_actual->quantum;
		  push(c,n_actual);
		  Cola_media();
		  Escribir(&count, &suma_lleg, &suma_long);
		  pop(c,n_actual);
		  cambios++;
		  Tiempos_espera2();
		}
	      }
	      else { /* le queda menos del quantum por terminar */
		t_actual = n->llegada;
		push(c,n);
		Cola_media();
		longitud++;
		Escribir(&count, &suma_lleg, &suma_long);
	      }
	    }
	    else { /* el ltimo llega despus de que acabe el actual */
	      if (n_actual->duracion > n_actual->quantum) {
		t_comienzo = t_actual += n_actual -> quantum;
		n_actual->duracion -= n_actual->quantum;
		n_actual->llegada = t_actual;
		n_actual->quantum = 2 * n_actual->quantum;
		push(c,n_actual);
		Cola_media();
		Escribir(&count, &suma_lleg, &suma_long);
		pop(c,n_actual);
		cambios++;
		Tiempos_espera2();
	      }
	      else {
		t_actual += n_actual->duracion;
		Escribir(&count, &suma_lleg, &suma_long);
		if (Cola_vacia(c)) {
		  *n_actual = *n;
		  cambios++;
		  t_actual = t_comienzo= n->llegada;
		  Escribir(&count, &suma_lleg, &suma_long);
		}
		else {
		  pop(c,n_actual);
		  cambios++;
		  Tiempos_espera2();
		  longitud--;
		  t_comienzo = t_actual;
		}
	      }
	    }
	  }
	} /* end else */
      } /* end else */
    } /* end else (proceso no dura ms de 1) */
  } /* end while (!feof(fichero)) */

  /* vaciamos la cola de preparados */
  while (!Cola_vacia(c)) {
    if (n_actual->duracion > n_actual->quantum) {
      t_actual = t_comienzo += n_actual->quantum;
      n_actual->duracion -= n_actual->quantum;
      n_actual->llegada = t_actual;
      n_actual->quantum = 2 * n_actual->quantum;
      push(c,n_actual);
      Cola_media();
      Escribir(&count, &suma_lleg, &suma_long);
      pop(c,n_actual);
      cambios++;
      Tiempos_espera2();
    }
    else {
      t_actual = t_comienzo+ n_actual->duracion;
      Escribir(&count, &suma_lleg, &suma_long);
      pop(c,n_actual);
      cambios++;
      Tiempos_espera2();
      t_comienzo = t_actual;
      longitud--;
    }
  }

   /* terminamos el ltimo */
  t_actual = t_comienzo + n_actual->duracion;
  Escribir(&count, &suma_lleg, &suma_long);


  Libera_cola(c);
  Analizar2();
  fclose(salida);
  fclose(salida2);
  fclose(salida3);
}

/***************************************************************************/
int main(int argc, char *argv[]) {

  if (argc != 2){
    printf("\nERROR, forma de uso: simul archivo de procesos");
    exit(1);
  }



  Abrir(argv[1]);	      /* apertura fichero de procesos */
  Iniciar();
  FCFS();
  RR();
  MRR();
  fclose(fichero);
  return 0;
}
